# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
    def self.get_file(dir_s4u_name,s4u_name,ext=".txt")
        file_name = s4u_name + ext
        file_name_full = File.join dir_s4u_name,file_name
        if File.exist?(file_name_full)
            file= File.new(file_name_full,"r")
            str=file.readlines
            file.close
            return str.collect{|s| s.chomp}
        end
        return false
        rescue=>error
            p error,error.backtrace.join('\n')
            return false
    end
    def self.set_file(dir_s4u_name,s4u_name,list,ext=".txt")
        file_name = s4u_name + ext
        file_name_full = File.join dir_s4u_name,file_name
        file= File.new(file_name_full, mode: 'w:UTF-8')
        str=list.join("\n")
        file.syswrite str
        file.close
        return true
        rescue=>error
            p error,error.backtrace.join('\n')
            return false
    end
	module S4u_panel
		class Panel
			def initialize(panel_mode=0)
				@panel_mode=panel_mode
				@golden_section=(1+Math.sqrt(5))/2
				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@snap=15*@scale_factor
                @panel_mode_list=[Strings["Panel by 3 points(corner)"],Strings["Panel by 3 points(middle)"],Strings["Panel by 3 points(center)"],Strings["Panel by face"]]
				@origin_list=[Strings["Back"],Strings["Center"],Strings["Front"]]
			end
			def activate
				model=Sketchup.active_model
				@edit_transform=Sketchup.active_model.axes.transformation
				@ip1=Sketchup::InputPoint.new
				@ip2=Sketchup::InputPoint.new
				@ip3=Sketchup::InputPoint.new
				@ip=Sketchup::InputPoint.new
				@faces=[]
				@element=[]

				self.get_setting()

				if @panel_mode==3
					faces=model.selection.to_a.select{|f| f.is_a?Sketchup::Face}
					if faces.length>0
						process_panel(faces)
					else
						#setting
					end
				else
					#setting
				end
				self.reset(model.active_view)
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip.position if @ip.valid?
				bb.add @ip1.position if @ip1.valid?
				bb.add @ip2.position if @ip2.valid?
				bb.add @ip3.position if @ip3.valid?
				return bb
			end
            def parse_setting(setting=nil)
                if setting
                    setting["offset"]= setting["offset"].collect{|e| parse_length(e)}
                    setting["thickness"] = parse_length(setting["thickness"])
                    setting["origin"] = setting["origin"].to_i
                    setting["make_group"] = setting["make_group"].to_i
                    setting["material"] = setting["material"].to_i
                    setting["delete_face"]= setting["delete_face"]
                    setting["snap_13"] = setting["snap_13"].to_f
                    setting["snap_center"] = setting["snap_center"].to_f
                    setting["max_recent"]=setting["max_recent"].to_i
                    setting["name"] = setting["name"].to_s
                    setting["layer"] = setting["layer"].to_s
                else#get default
                    setting={}
                    setting["offset"]=[0,0,0,0]
                    setting["thickness"]=10.cm
                    setting["origin"]=0
                    setting["make_group"]=0
                    setting["material"]=0
                    setting["delete_face"]=false
                    setting["snap_13"]=0
                    setting["snap_center"]=@snap
                    setting["max_recent"]=8
                    setting["name"]=Strings["Panel"]
                    setting["layer"]= Sketchup.active_model.active_layer.name
                    if Sketchup.version.to_i>=20
                        setting["layer"]= Sketchup.active_model.layers[0].display_name
                    end
                end
                return setting
            end
            def get_setting()
                @last_setting = Module.nesting[-2].instance_variable_get :@last_setting
                if !(@last_setting && @last_setting[0].class==Hash)
                    @last_setting=[] if !@last_setting
                    if s=S4U::get_file(Dir_s4u_name,NAME)
                        s.compact!
                        (0..s.length-1).each{|i| @last_setting << self.parse_setting(JSON.parse(s[i])) if s[i]}
                    else
                        setting=self.parse_setting()
                        @last_setting.unshift(setting) if @last_setting.empty? ||
                          (@last_setting[0] && @last_setting[0].keys.any?{|e| @last_setting[0][e]!=setting[e]})
                    end
                end
                @setting=@last_setting[0]
                @last_setting=@last_setting[0..@setting['max_recent']]
                rescue=>error
                @setting=self.parse_setting
                @last_setting=[] if !@last_setting
                @last_setting[0]=@setting
                p error,error.backtrace.join("\n")
            end
            def save_setting
                setting=@setting
                @last_setting.unshift(setting) if @last_setting.empty? ||
                     (@last_setting[0] && @last_setting[0].keys.any?{|e| @last_setting[0][e]!=setting[e]})
                Module.nesting[-2].instance_variable_set :@last_setting,@last_setting
                S4U::set_file(Dir_s4u_name,NAME, @last_setting.collect{|s| s.to_json}[0..setting['max_recent']].compact)
                rescue=>error
                p error,error.backtrace.join("\n")
            end

			def deactivate(view)
                self.save_setting
            end
			def setstatus(view=nil)
				text = Strings["Panel Mode"] + ": "
				text+= @panel_mode_list[@panel_mode].upcase
				text+=" (" + Strings["RightMouse for setting"] + ") " + " | "
				text+= Strings["Origin Panel"] + ": "
				text+= @origin_list[@setting["origin"]].upcase
				text+=" (" + Strings["Press TAB for change"] + ")| "
				if @panel_mode<3
					if @state==0
						text+=Strings["Select First point"]
					elsif @state==1
						text+= Strings["Select Second point"]
					elsif @state==2
						text+= Strings["Select Third point"]
					end
				elsif @panel_mode==3
					text+=Strings["Select Face"]
				end
				Sketchup.status_text=text
				view.invalidate if view
			end
			def getMenu(menu)
				menus=[]
				@panel_mode_list.each_index{|i| menus<< menu.add_item(@panel_mode_list[i]) {@panel_mode=i; self.setstatus }}
				menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
				menu.set_validation_proc(menus[@panel_mode]) { MF_CHECKED }
				menu.add_separator
				menus=[]
				@origin_list.each_index{|i| menus<< menu.add_item(@origin_list[i]) {@setting["origin"]=i; self.setstatus }}
				menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
				menu.set_validation_proc(menus[@setting["origin"]]) { MF_CHECKED }
				menu.add_separator
				menus=[]
				menus<< menu.add_item(Strings["Set Parameter"]) { self.setting ; self.setstatus }
			end
			def draw(view)
				model=view.model
				view.line_width=2
				view.drawing_color="black"
				if @panel_mode<3
					if @ip && @ip.valid?
						if @ip.display?
							@ip.draw(view)
						elsif @face || @edge
							view.drawing_color="red"
							(view.draw_points @ip.position, 9*@scale_factor, 1, "red" )
						end
					end
					if @state==1
						if @ip1.valid?	&& @ip2.valid?
							@ip1.draw(view)
							view.set_color_from_line @ip1.position,@ip2.position
							view.draw_line @ip1.position,@ip2.position
							if @ip_shift && @shift_press
								view.line_stipple = "."
								view.drawing_color ="blue"
								view.draw_line(@ip.position, @ip2.position)
								view.line_stipple = ""
								view.drawing_color ="red"
								p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
								view.draw_points p2_line, 9*@scale_factor, 2, "red"  if p2_line
							end
						end
					elsif @state==2
						@ip1.draw(view)
						@ip2.draw(view)
						if @ip3 && @ip3.valid?
							@ip3.draw(view)
							if @point
								view.draw_points @p3, 9*@scale_factor, 1, "red"
								view.line_stipple = "."
								if @panel_mode==0
									view.draw_line @p1,@p3
								elsif @panel_mode==1
									view.draw_line( (@p1+(@p1-@p2)),@p3)
								elsif @panel_mode==2
									view.draw_line( (@p1+(@p1-@p3)),@p3)
								end
								view.line_stipple = ""
							end
							ps=get_rectangle([@ip1.position,@ip2.position,@p3])
							if ps && ps.length==4
								ps1=get_points_panel(ps)
								if ps1
									view.drawing_color="red"
									view.draw_polyline (ps1[0]<<ps1[0][0])
									if @setting["thickness"]>0
										view.draw_polyline (ps1[1]<<ps1[1][0])
										view.draw_line ps1[2]
									end
								end
							end
						else
							view.draw_line @ip1.position,@ip2.position
						end
					end
				elsif @panel_mode==3 && @face_points
					ps=@face_points[0].uniq
					ps=ps.collect{|p| p.clone}
					if ps && ps.length==4
						ps1=get_points_panel(ps)
						if ps1
							view.drawing_color="red"
							view.draw_polyline (ps1[0]<<ps1[0][0])
							if @setting["thickness"]>0
								view.draw_polyline (ps1[1]<<ps1[1][0])
								view.draw_line ps1[2]
							end
						end
					end
				end
				self.setstatus
			end
			def get_rectangle(points)
				ps=[]
				if points.length==3
					if @panel_mode==0
						ps=points
						vx=points[1]-points[0]
						p2_line=points[2].project_to_line([points[0],vx])
						vyt=points[2]-p2_line
						ps[2]=points[1] + vyt
						vy=ps[2]-points[1]
						ps[3]=points[0] + vy
						return ps
					elsif @panel_mode==1
						vy=points[1]-points[0]
						p3_line=points[2].project_to_line([points[0],vy])
						vx=points[2]-p3_line
						ps[0]=points[0] + vy.reverse
						ps[1]=points[1]
						ps[2]=points[1] + vx
						ps[3]=ps[0] + vx
						return ps
					elsif @panel_mode==2
						vy=points[1]-points[0]
						p3_line=points[2].project_to_line([points[0],vy])
						vxt=points[2]-p3_line
						ps[3]=points[1] + vxt
						vx=points[1]-ps[3]
						ps[0]=points[0] + (vx.reverse+vy.reverse)
						ps[1]=points[0] + (vx+vy.reverse)
						ps[2]=points[1] + vx
						return ps
					end
				end
				return nil
			end
			def get_points_panel(points)
				vx=(points[1]-points[0])
				vy=(points[3]-points[0])
				vz=vx.cross vy
				points[3].offset! vx.reverse,@setting["offset"][0]
				points[0].offset! vx.reverse,@setting["offset"][0]
				points[0].offset! vy.reverse,@setting["offset"][3]
				points[1].offset! vy.reverse,@setting["offset"][3]
				points[1].offset! vx,@setting["offset"][1]
				points[2].offset! vx,@setting["offset"][1]
				points[2].offset! vy,@setting["offset"][2]
				points[3].offset! vy,@setting["offset"][2]
				vz.length=@setting["thickness"] if @setting["thickness"]!=0
				points1=[];points2=[]
				points.each_index{|i| points1[i]=points[i]+vz }
				depth=-@setting["thickness"]*0.5*@setting["origin"]
				if depth!=0
					vec=vz
					vec.length=depth
					points.collect!{|p| p+vec}
					points1.collect!{|p| p+vec}
				end
				points.each_index{|i| points2.push points[i],points1[i]}
				return [points,points1,points2]
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@setting["snap_13"]) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip= Strings["1/3"]
					return ip
				end
				return nil
			end
			def set_special_point(x,y,view,ip)
				if ip.face && (@setting["snap_center"]>0)
					cf=center_face(ip.face).transform ip.transformation
					ph = view.pick_helper
					ph.init(x, y)
					picked = ph.test_point(cf,x,y,@setting["snap_center"])
					if picked
						ip1=Sketchup::InputPoint.new(cf)
						@face=true
						view.tooltip= Strings["Center face"]
						ip.copy! ip1
					elsif (@setting["snap_13"]>0)
						@edge=ip.face.edges.find{|e| set_pointonedge(x,y,view,ip,e) }
					end
				elsif ip.edge && (@setting["snap_13"]>0)
					@edge=set_pointonedge(x,y,view,ip,ip.edge)
				end
			end
			def find_special_point(x,y,view,p1,p2,ip,d1)
				if p1!=p2
					d1*=2 if @box_mode==1
					ph = view.pick_helper
					vec=(p2-p1)
					vec.length=d1
					p_square= p1 + vec
					screen_square=view.screen_coords p_square
					vec.length=d1*@golden_section
					p_golden1= p1 + vec
					screen_golden1=view.screen_coords p_golden1
					vec.length=d1/@golden_section
					p_golden2= p1 + vec
					screen_golden2=view.screen_coords p_golden2
					p=p2
					if ph.init(p_square.x, p_square.y) && ph.test_point(p2,screen_square.x, screen_square.y,(@snap-5))
						p2=p_square
						view.tooltip= Strings["Square"]
					elsif ph.init(screen_golden1.x, screen_golden1.y) && ph.test_point(p2,screen_golden1.x, screen_golden1.y,(@snap-5))
						p2=p_golden1
						view.tooltip= Strings["Golden section"]
					elsif ph.init(screen_golden2.x, screen_golden2.y) && ph.test_point(p2,screen_golden2.x, screen_golden2.y,(@snap-5))
						p2=p_golden2
						view.tooltip= Strings["Golden section"]
					end
					if p!=p2
						ip.copy! Sketchup::InputPoint.new(p2)
						Sketchup::set_status_text( (p1.distance p2).to_s, SB_VCB_VALUE)
						return ip
					end
				end
			end
			def onMouseMove(flags, x, y, view)
				@face=@edge=@point=nil
				if @panel_mode<3
					@face=nil
					if @state==0
						@ip.pick view,x,y
						set_special_point(x,y,view,@ip)
						view.invalidate if @ip.valid?
					elsif @state==1
						@ip.pick view,x,y,@ip1
						set_special_point(x,y,view,@ip)
						if @shift_press && @ip_shift && @ip!=@ip1 && @ip2!=@ip1 && (@ip1.position!=@ip_shift)
							p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
							@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
						else
							@ip2.copy! @ip
						end
						@p2=@ip2.position
						view.invalidate if @ip.valid?
						length = @ip1.position.distance(@ip2.position)
						Sketchup::set_status_text( length.to_s, SB_VCB_VALUE)
					elsif @state==2
						@ip.pick view,x,y,@ip2
						set_special_point(x,y,view,@ip)
						if @ip.valid? && ((@ip.position.distance_to_line [@p1,@p2])>0)
							if @shift_press	&& @ip3.valid? && ((@ip3.position.distance_to_line [@p1,@p2])>0)
								p2_line=@ip3.position.project_to_line([@p1,@p2])
								vyt=@ip3.position-p2_line
								p3_line=@ip.position.project_to_line([@p2,vyt])
								vy=p3_line-@p2
								@p3=@p2 + vy
							else
								p2_line=@ip.position.project_to_line([@p1,@p2])
								vy=@ip.position-p2_line
								@p3=@p2 + vy
							end
							@ip3.copy! (Sketchup::InputPoint.new(@p3))
							view.invalidate if @ip3.valid?
							d=@p2.distance(@p1)
							ip=find_special_point(x,y,view,@p2,@p3,@ip,d)
							if ip
								@ip3.copy! ip
								@p3=@point=ip.position
								view.invalidate if @ip3.valid?
							end
							length = @ip2.position.distance(@ip3.position)
							Sketchup::set_status_text( length.to_s, SB_VCB_VALUE)
						end
					end
				elsif @panel_mode==3
					@ip.pick view,x,y
					ph = view.pick_helper
					ph.do_pick(x, y)
					vertices=nil
					vertices=get_vertices(@ip.face)
					@face_panel=nil
					@face_points=nil
					@instance_path=nil
					if vertices && vertices.length==4
						face=vertices.collect{|v| v.position.transform @ip.transformation}
						if Geom.point_in_polygon_2D(@ip.position,face,true)
							@face_panel=@ip.face
							@instance_path= @ip.instance_path
							@face_points=[face,@ip.face]
						end
					end
					view.invalidate
				end
			end
			def onLButtonDown(flags, x, y, view)
				if @panel_mode<3
					if @state==0
						if @ip.valid?
							@element.clear
							@ip1.copy! @ip
							@p1=@ip1.position
							@state=1
							view.lock_inference if  view.inference_locked?
						end
					elsif @state==1 && @ip.valid? && (@ip1.position!=@ip.position)
						#@ip2.copy! @ip
						@state=2
						#@p2=@ip2.position
						view.lock_inference if  view.inference_locked?
					elsif @state==2
						self.reset(view)
						process_panel([@p1,@p2,@p3])
					end
				elsif @panel_mode==3
					if @face_points && !(@faces.include?(@face_points))
						@element.clear
						@faces<<@face_points
						process_panel(nil)
					end
				end
				self.setstatus(view)
			end
			def setting
                @dialog= Module.nesting[-2].instance_variable_get(:@dialog)
                if !@dialog
                    @dialog = UI::HtmlDialog.new(
                        {
                            :dialog_title => Strings["Setting"],
                            :preferences_key => "suforyou.s4u_panel",
                            :resizable => true,
                            :scrollable => true,
                            :width => 400,
                            :height => 495,
                            :min_width => 150,
                            :min_height => 350,
                            :style => UI::HtmlDialog::STYLE_DIALOG
                        }
                    )
                    @dialog.center
                end
                @dialog.add_action_callback("apply") { |action_context,setting,last_setting|
                    model=Sketchup.active_model
                    layers=model.layers.collect{|l| l.name}
                    if Sketchup.version.to_i>=20
                        layers[0]=model.layers[0].display_name
                    end
                    setting=JSON.parse setting
                    @last_setting=[] if !@last_setting
                    setting=self.parse_setting(setting)
                    @last_setting=@last_setting[0..setting['max_recent']]
                    if @last_setting.empty? || @last_setting[0].keys.any?{|e| @last_setting[0][e]!=setting[e]}
                        @last_setting.unshift(setting)
                        @dialog.execute_script("last_setting=#{@last_setting.collect{|e| e.to_json}};
                                                layers=#{layers};
                                                load(last_setting,layers);
                                                ")
                    end
                    @setting=@last_setting[0]
                    self.save_setting
                }
                @dialog.add_action_callback("load") { |action_context,response|
                    model=Sketchup.active_model
                    layers=model.layers.collect{|l| l.name}
                    if Sketchup.version.to_i>=20
                        layers[0]=model.layers[0].display_name
                    end
                    if @last_setting
                        @dialog.execute_script("
                                                last_setting=#{@last_setting.collect{|e| e.to_json}};
                                                layers=#{layers};
                                                load(last_setting,layers);
                                                ")
                    end
                }
                @dialog.add_action_callback("reset") { |action_context,response|
                    model=Sketchup.active_model
                    layers=model.layers.collect{|l| l.name}
                    if Sketchup.version.to_i>=20
                        layers[0]=model.layers[0].display_name
                    end
                    if @last_setting
                        @dialog.execute_script("
                                                last_setting=#{@last_setting.collect{|e| e.to_json}};
                                                layers=#{layers};
                                                load(last_setting,layers);
                                                ")
                    end
                }
                @dialog.set_html S4U::S4u_panel.get_html
                Module.nesting[-2].instance_variable_set(:@dialog,@dialog)
                @dialog.show
			end
			def cal(input)
				lengthformat=Sketchup.active_model.options["UnitsOptions"]["LengthFormat"]
				if op=input[/[\+\-\/\*]/]
					begin
					if input[/[\+\-]/]
						index= input.rindex(/[\+\-]/)
					elsif input[/[\*\/]/]
						index= input.rindex(/[\*\/]/)
					end
					if index
						if index>0
							s=[input[0..(index-1)],input[(index+1)..-1]]
						else
							s=["0",input[(index+1)..-1]]
						end
						op=input[index]
						if (lengthformat==0 || lengthformat==2)
							if op=="+"
								value= cal(s[0]).to_s.to_f + cal(s[1]).to_s.to_f
							elsif op=="-"
								value= cal(s[0]).to_s.to_f - cal(s[1]).to_s.to_f
							elsif op=="*"
								value= cal(s[0]).to_s.to_f * cal(s[1]).to_s.to_f
							elsif op=="/"
								value= cal(s[0]).to_s.to_f / cal(s[1]).to_s.to_f
							end
						else
							if op=="+"
								value= cal(s[0]).to_f + cal(s[1]).to_f
							elsif op=="-"
								value= cal(s[0]).to_f - cal(s[1]).to_f
							elsif op=="*"
								value= cal(s[0]).to_f * cal(s[1]).to_f
							elsif op=="/"
								value= cal(s[0]).to_f / cal(s[1]).to_f
							end
						end
					end
					rescue
						value=parse_length(input)
					end
				else
					value = parse_length(input)
				end
				return parse_length(value)
			end
			def parse_length(input)
				(return 0) if !input
				input=input.to_s
				if !Sketchup.parse_length("1.1")
					input.gsub!(/\./, ',')
				else
					input.gsub!(/\,/, '.')
				end
				begin
					value=input.to_l
				rescue
					value=input.to_f
				end
				return value
			end
			def onReturn(view)
				view.model.select_tool nil
			end
			def process_panel(points=[])
				model=Sketchup.active_model
				model.start_operation TOOL,true,false,false
					if @panel_mode<3
						ps=get_rectangle(points)
						@element<< make_panel(ps)
					elsif @panel_mode==3
						points.each{|f|
							vertices=get_vertices(f)
							if vertices && vertices.length==4
								(@faces<< [vertices.collect{|v| v.position},f])
								if @setting["delete_face"] && f && f.valid? && f.is_a?(Sketchup::Face)
									ed=f.edges
									f.erase! if f && f.valid?
									ed.each{|e| e.erase! if e && e.valid? && !e.deleted? && e.faces.length==0 }
								end
							end
						} if points
						@faces.each{|ps1| @element<< make_panel(ps1[0],ps1[1]) if ps1 }
						if @setting["delete_face"] && @face_panel && @face_panel.valid?  && @face_panel.is_a?(Sketchup::Face)
							ed=@face_panel.edges
							@face_panel.erase! if @face_panel && @face_panel.valid? && !@face_panel.deleted?
							ed.each{|e| e.erase! if e && e.valid? && !e.deleted? && e.faces.length==0 }
						end
					end
				model.commit_operation
				self.reset(model.active_view)
				self.setstatus
				rescue=>error
				p error,error.backtrace.join("\n")
				model.abort_operation
			end
			def make_panel(points,face=nil)
				return if points && points.uniq.length!=4
				model=Sketchup.active_model
				entities=model.active_entities
				#materials=model.materials.collect{|m| m.name}
				material_current=model.materials.current
				vx=(points[1]-points[0])
				vy=(points[3]-points[0])
				vz=vx.cross vy
				g=entities.add_group
				g.transformation=Geom::Transformation.new(vx,vy,vz,points[0])
				points[3].offset! vx.reverse,@setting["offset"][0]
				points[0].offset! vx.reverse,@setting["offset"][0]
				points[0].offset! vy.reverse,@setting["offset"][3]
				points[1].offset! vy.reverse,@setting["offset"][3]
				points[1].offset! vx,@setting["offset"][1]
				points[2].offset! vx,@setting["offset"][1]
				points[2].offset! vy,@setting["offset"][2]
				points[3].offset! vy,@setting["offset"][2]
				ps=points.collect{|p| p.transform g.transformation.inverse}
				f=g.entities.add_face ps
				if f
					f.reverse!
					f.pushpull(@setting["thickness"], false) if @setting["thickness"]!=0
					depth=-@setting["thickness"]*0.5*@setting["origin"]
					if depth!=0
						vec=g.transformation.zaxis
						vec.length=depth
						g.transformation=(Geom::Transformation.translation(vec))*g.transformation
					end
				else
					return nil
				end
				if @setting["name"].to_s!=""
					if @setting["make_group"]==1
						g=g.to_component
						(g.definition.name=g.name=@setting["name"].to_s)
					elsif @setting["make_group"]==0
						if Sketchup.version.to_i>=15
							g.definition.name=g.name=@setting["name"].to_s
						else
							g.name=@setting["name"].to_s
						end
					end
				end
				if @setting["material"]==1
					g.material=material_current if material_current
				elsif @setting["material"]==2 && face && face.is_a?(Sketchup::Face)
					mat=face.material ? face.material : face.back_material
					if !mat
						path=@instance_path.to_a.reverse.select{|e| e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)}
						path.each{|e|
							if e.material
								mat=e.material
								break
							end
						}
					end
					g.material=mat
				end
				# if @setting["material"] != -1
					# if @setting["material"]<(model.materials.length) && model.materials.length>0
						# g.material=model.materials[@setting["material"]]
					# end
				# elsif model.materials.current
					# g.material=model.materials.current
				# end
				g.layer=model.layers[@setting["layer"]]
				return g
			end
			def get_vertices(face)
				return nil if not( face && (face.is_a?Sketchup::Face))
				ver=[]
				vertices=[]
				vertices =face.outer_loop.vertices*3
				(0..face.outer_loop.vertices.length-1).each { |i|
					j=face.outer_loop.vertices.length+i
					p1=vertices[j-1].position.clone
					p2=vertices[j].position.clone
					p3=vertices[j+1].position.clone
					v1=p1.vector_to p2
					v2=p2.vector_to p3
					ver.push vertices[j] if not( v1.samedirection? v2)
				}
				return ver
			end
			def onKeyDown(key, repeat, flags, view)
				if key==CONSTRAIN_MODIFIER_KEY
					@shift_press=true
					if view.inference_locked?
						view.lock_inference
					elsif (@state==0 && @ip.valid? )
						view.lock_inference @ip
					elsif (@state==1 && @ip.valid?)
						@ip_shift=@ip.position
						view.lock_inference @ip,@ip1
					elsif @state==2 && @ip3.valid?
						@ip_shift= @ip3
						view.lock_inference @ip3,@ip2
					end
				elsif key==VK_UP
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p= ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.zaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				elsif key==VK_LEFT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p=  ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.yaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				elsif key==VK_RIGHT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p= ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.xaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				end
				self.setstatus(view)
			end
			def onKeyUp(key, repeat, flags, view)
				if (key==CONSTRAIN_MODIFIER_KEY)
					@shift_press=false
					view.lock_inference if view.inference_locked?
				elsif ((key==9) && IS_WIN ) || ((key==48) && IS_OSX)
					origin= (@setting["origin"]==0) ? 2 : (@setting["origin"]-1)
					@panel_mode<3 ? set_origin(origin) :(@setting["origin"]=origin)
				elsif key==VK_DOWN
					origin= (@setting["origin"]==0) ? 2 : (@setting["origin"]-1)
					set_origin(origin)
				end
				self.setstatus(view)
			end
			def onUserText(text, view)
				if @state==1
					if @ip.position!=@p1
						vec=@ip.position - @p1
						length=self.cal(text)
						Sketchup::set_status_text( length.to_s, SB_VCB_VALUE)
						if length!=0 && vec.length>0
							vec.length=length
							@p2=@p1 + vec
							ip = Sketchup::InputPoint.new(@p2)
							@ip2.copy! ip
							@state=2
							view.lock_inference if  view.inference_locked?
						end
					end
				elsif @state==2
					if @ip.position!=@p2
						vx=@p2-@p1
						p2_line=@ip.position.project_to_line([@p1,vx])
						vy=@ip.position-p2_line
						length=self.cal(text)
						Sketchup::set_status_text( length.to_s, SB_VCB_VALUE)
						if length!=0 && vy.length>0
							vy.length=length
							@p3=@p2 + vy
							process_panel([@p1,@p2,@p3])
						end
					end
				end
				self.setstatus(view)
			end
			def center_face(face)
				vertices=get_vertices(face)
				return nil if !vertices
				numver=vertices.length
				cp=Geom::Point3d.new
				vertices.each{|v|
					cp.x=cp.x+v.position.x
					cp.y=cp.y+v.position.y
					cp.z=cp.z+v.position.z
				}
				cp.x=cp.x/ numver
				cp.y=cp.y/ numver
				cp.z=cp.z/ numver
				return cp
			end
			def set_origin(origin)
				depth=@setting["thickness"]*0.5*(@setting["origin"]-origin)
				if depth!=0 && @element && @element.length>0
					model=Sketchup.active_model
					model.start_operation TOOL,true,false,false
					@element.each{|e|
						vec=e.transformation.zaxis
						vec.length=depth
						e.transformation=(Geom::Transformation.translation(vec))*e.transformation
					}
					model.commit_operation
				end
				@setting["origin"]=origin
			end
			def onCancel(reason, view)
				self.reset(view)
			end
			def reset(view=nil)
				@state=0
				@multiselect=0
				@faces.clear
				@face=nil
				@ip1.clear
				@ip2.clear
				@ip3.clear
				@ip.clear
				view.lock_inference if view && view.inference_locked?
				self.setstatus(view)
			end
			def resume(view)
				self.setstatus(view)
			end
		end
		def self.main(panel_mode=0)
			@panel_mode=0 if !@panel_mode
			@panel_mode=panel_mode if panel_mode<4
			Sketchup.active_model.select_tool Panel.new(@panel_mode)
		end
        def self.get_html
            html=
            <<EOS
                <!DOCTYPE html>
                <html>
                    <style>
                        .input-group-text{
                            width: 100px;
                        }
                        .btn {
                            width: 80px;
                        }
                    </style>
                    <link href='#{Dir_s4u_name}/vendor/modus/modus.min.css' rel='stylesheet'>
                    <script src='#{Dir_s4u_name}/vendor/modus/jquery.js'></script>
                    <script src='#{Dir_s4u_name}/vendor/modus/bootstrap.bundle.min.js'></script>
                    <script>
                        function context(event) {
                            event.preventDefault();
                        }
                        function keyright(event) {
                            if(event.shiftKey && (event.key=='F10' || event.keyCode==121 ))
                            {
                                event.preventDefault();
                            }
                        }
                    </script>
               <!-- <body oncontextmenu='context(event)' onkeydown='keyright(event)' class='m-3 bg-panel-background'> -->
                    <body class='m-3 bg-panel-background'>
                        <div class="container p-0" id='dlg-content-area'>
                            <div class='input-group mb-2'  >
                                <div class='input-group-prepend '>
                                    <span class='input-group-text bg-panel-background' id='text_setting'>#{Strings['Setting Recent']}</span>
                                </div>
                                <select class='custom-select form-control' id='setting' onchange='get_setting()'>

                                </select>
                            </div>
                            <div class="form-group mb-2">
                                <div class='input-group'>
                                    <div class='input-group-prepend'>
                                        <span class='input-group-text bg-panel-background' id='text_offset1'>#{Strings['Left']}</span>
                                    </div>
                                    <input class='form-control' id='offset1'>
                                </div>
                                <div class='input-group'>
                                    <div class='input-group-prepend'>
                                        <span class='input-group-text bg-panel-background' id='text_offset2'>#{Strings['Right']}</span>
                                    </div>
                                    <input class='form-control' id='offset2' >
                                </div>
                                <div class='input-group'>
                                    <div class='input-group-prepend'>
                                        <span class='input-group-text bg-panel-background' id='text_offset3'>#{Strings['Top']}</span>
                                    </div>
                                    <input class='form-control' id='offset3'>
                                </div>
                                <div class='input-group'>
                                    <div class='input-group-prepend'>
                                        <span class='input-group-text bg-panel-background' id='text_offset4'>#{Strings['Bottom']}</span>
                                    </div>
                                    <input class='form-control' id='offset4'>
                                </div>
                                <div class='input-group'>
                                    <div class='input-group-prepend'>
                                        <span class='input-group-text bg-panel-background' id='text_thickness'>#{Strings['Thickness']}</span>
                                    </div>
                                    <input class='form-control' id='thickness'>
                                </div>
                            </div>

                            <div class='input-group'>
                                <div class='input-group-prepend'>
                                    <span class='input-group-text bg-panel-background' id='text_Name'>#{Strings['Name']}</span>
                                </div>
                                <input class='form-control' id='name'>
                            </div>
                            <div class='input-group'>
                                <div class='input-group-prepend'>
                                    <span class='input-group-text bg-panel-background' id='text_tags'>#{Strings['Tags']}</span>
                                </div>
                                <select class='custom-select form-control' id='tags'>

                                </select>
                            </div>
                            <div class='input-group'>
                                <div class='input-group-prepend'>
                                    <span class='input-group-text bg-panel-background' id='text_material'>#{Strings['Material']}</span>
                                </div>
                                <select class='custom-select form-control' id='material'>
                                    <option>#{Strings['Default']}</option>
                                    <option>#{Strings['Current']}</option>
                                    <option>#{Strings['Face']}</option>
                                </select>
                            </div>
                            <div class='input-group'>
                                <div class='input-group-prepend'>
                                    <span class='input-group-text bg-panel-background' id='text_make_group'>#{Strings['Make Group']}</span>
                                </div>
                                <select class='custom-select form-control' id='make_group'>
                                    <option>#{Strings['Group']}</option>
                                    <option>#{Strings['Component']}</option>
                                </select>
                            </div>
                        </div>
                        <div class="accordion accordion-sm" id="more_setting">
                            <div class="card">
                                <div class="card-header" id="more_setting_header1" data-toggle="collapse" data-target="#more_setting_data1" aria-expanded="false" aria-controls="codeCollapseOne">
                                    <span class="mb-0" id="collapsible-group-item-1">
                                        More Setting
                                    </span>
                                </div>
                                <div id="more_setting_data1" class="collapse" aria-labelledby="more_setting_header1">
                                    <div class="card-body">
                                        <div class='input-group'>
                                            <div class='input-group-prepend'>
                                                <span class='input-group-text bg-panel-background' id='text_snap_13'>#{Strings['Snap 1/3']}</span>
                                            </div>
                                            <input class='form-control' id='snap_13'>
                                        </div>
                                        <div class='input-group'>
                                            <div class='input-group-prepend'>
                                                <span class='input-group-text bg-panel-background' id='text_snap_center'>#{Strings['Snap Center']}</span>
                                            </div>
                                            <input class='form-control' id='snap_center'>
                                        </div>
                                        <div class='input-group'>
                                            <div class='input-group-prepend'>
                                                <span class='input-group-text bg-panel-background' id='text_max_recent'>#{Strings['Max Recent']}</span>
                                            </div>
                                            <input class='form-control' id='max_recent'>
                                        </div>
                                        <div class='custom-control custom-switch mt-2 '>
                                            <input type='checkbox' class='custom-control-input' id='delete_face'>
                                            <label class='custom-control-label' for='delete_face'>
                                                #{Strings['Delete Face']}
                                            </label>
                                        </div>
                                    </div>
                                </div>
                            </div>
                        </div>
                        <div class="container p-0 mb-0">
                            <div class="modal-footer pr-0 pb-0">
                                <button type='button' class='btn btn-outline-dark mb-0' onclick='sketchup.reset()' >
                                    #{Strings['Reset']}
                                </button>
                                <button type='button' class='btn btn-outline-dark mb-0' onclick='window.close()' >
                                    #{Strings['Cancel']}
                                </button>
                                <button type='button' class='btn btn-primary mr-0 mb-0' onclick='apply()'>
                                    #{Strings['Apply']}
                                </button>

                            </div>
                        </div>
                    <script>
                    var last_setting=[];
                    function load(last_setting ,layers){
                        for (let i = 0; i < last_setting.length; i++) {
                            last_setting[i]= JSON.parse(last_setting[i]);
                        }

                        var parameter=['offset1','offset2','offset3','offset4'];
                        for (let i = 0; i < parameter.length; i++) {
                            document.getElementById(parameter[i]).value=last_setting[0]['offset'][i];
                        }
                        document.getElementById('thickness').value= last_setting[0]['thickness'];

                        document.getElementById('name').value= last_setting[0]['name'];
                        document.getElementById('snap_13').value= last_setting[0]['snap_13'] + ' pixels';
                        document.getElementById('snap_center').value= last_setting[0]['snap_center'] + ' pixels';
                        document.getElementById('max_recent').value= Number(last_setting[0]['max_recent'])+1;
                        document.getElementById('delete_face').checked= last_setting[0]['delete_face'];

                        var materials=['Default','Current','Face'];
                        document.getElementById('material').value= materials[Number(last_setting[0]['material'])];
                        var make_group=['Group','Component'];
                        document.getElementById('make_group').value= make_group[Number(last_setting[0]['make_group'])];

                        list=document.getElementById('setting');
                        while (list.hasChildNodes()) {
                            list.removeChild(list.firstChild);
                        }
                        for (let j = 0; j < last_setting.length; j++) {
                            const option = document.createElement('option');
                            text='S'+ (j+1) + ': '
                            for (let i = 0; i < parameter.length; i++) {
                                text+= i>0 ? '|' : ''
                                text+=  last_setting[j]['offset'][i];
                            }
                            text+=  '|' + last_setting[j]['thickness'];
                            const node = document.createTextNode(text);
                            option.appendChild(node);
                            list.appendChild(option);
                        }

                        list=document.getElementById('tags');
                        while (list.hasChildNodes()) {
                            list.removeChild(list.firstChild);
                        }
                        for (let j = 0; j < layers.length; j++) {
                            const option = document.createElement('option');
                            const node = document.createTextNode(layers[j]);
                            option.appendChild(node);
                            list.appendChild(option);
                        }
                        for (let j = 0; j < layers.length; j++) {
                            if( layers[j]==last_setting[0]['layer'] )
                            {
                                document.getElementById('tags').value=last_setting[0]['layer'];
                            }
                        }
                    }
                    function get_setting(){
                        list=document.getElementById('setting').value;
                        id= list.split(':')[0].slice(1);
                        id=Number(id)-1;
                        var materials=['Default','Current','Face'];
                        var make_group=['Group','Component'];

                        var ids=['offset1','offset2','offset3','offset4'];
                        for (let i = 0; i < ids.length; i++) {
                            document.getElementById(ids[i]).value=last_setting[id]['offset'][i];
                        }
                        document.getElementById('thickness').value= last_setting[id]['thickness'];
                        document.getElementById('name').value= last_setting[id]['name'];
                        document.getElementById('snap_13').value= last_setting[id]['snap_13'] + ' pixels';
                        document.getElementById('snap_center').value= last_setting[id]['snap_center'] + ' pixels';
                        document.getElementById('max_recent').value= Number(last_setting[id]['max_recent']) +1;
                        document.getElementById('delete_face').checked= last_setting[id]['delete_face'];

                        document.getElementById('material').value= materials[Number(last_setting[id]['material'])];
                        document.getElementById('make_group').value= make_group[Number(last_setting[id]['make_group'])];

                        for (let j = 0; j < layers.length; j++) {
                            if( layers[j]==last_setting[0]['layer'] )
                            {
                                document.getElementById('tags').value=last_setting[id]['layer'];
                            }
                        }
                    }
                    function apply()
                    {
                        var arr={};
                        var ids=['offset1','offset2','offset3','offset4'];
                        arr['offset']=[];
                        for (let i = 0; i < 4; i++) {
                            arr['offset'][i]=document.getElementById(ids[i]).value;
                        }
                        arr['thickness']= $('#thickness').val();
                        arr['name']= $('#name').val();
                        var materials_hash={'Default':0,'Current':1,'Face':2}
                        arr['material']= materials_hash[$('#material').val()];
                        var make_group_hash={'Group':0,'Component':1};
                        arr['make_group']= make_group_hash[$('#make_group').val()];
                        arr['layer']= $('#tags').val();
                        arr['snap_13']= $('#snap_13').val();
                        arr['snap_center']= $('#snap_center').val();
                        arr['delete_face']= $('#delete_face').prop('checked');
                        max=Number($('#max_recent').val())
                        arr['max_recent']= max>0 ? (max-1) : 0;
                        console.log(arr);
                        sketchup.apply(JSON.stringify(arr),last_setting);
                    }
                    sketchup.load();
                </script>
                </body>
                </html>
EOS
            end
	end
end